# -*- coding: utf-8 -*-
u""" The :func:`~ftr.process.ftr_process` function makes use of all FTR
objects to offer an easy-to-use *one-liner* to integrate in your python
code.

:func:`~ftr.process.ftr_process` is suitable for live extraction (content
currently available on the internet), but also for postponed or post-mortem
extraction where the content was removed from the internet but you still
have the HTML and the original URL handy.

.. note:: as of current version the :func:`~ftr.process.ftr_process`
    wrapper is the only way to get multiple-page articles parsed as a
    whole (eg. all pages extracted, cleaned and appended as one). See
    :class:`~ftr.extractor.ContentExtractor` for details.

.. Copyright 2015 Olivier Cortès <oc@1flow.io>.

    This file is part of the python-ftr project.

    python-ftr is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    python-ftr is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public
    License along with python-ftr. If not, see http://www.gnu.org/licenses/
"""
import logging

try:
    import requests
except ImportError:
    # Happens during installation before setup.py finishes installing deps.
    requests = None

from .config import ftr_get_config, SiteConfig, CACHE_TIMEOUT, cached
from .extractor import ContentExtractor

try:
    from sparks.utils.http import detect_encoding_from_requests_response

except ImportError:
    # same problem, same effect.
    pass

LOGGER = logging.getLogger(__name__)


@cached(timeout=CACHE_TIMEOUT)
def requests_get(url):
    """ Run :func:`requests.get` in a ``cached()`` wrapper.

    The cache wrapper uses the default timeout (environment variable
    ``PYTHON_FTR_CACHE_TIMEOUT``, 3 days by default).

    It is used in :func:`ftr_process`.
    """

    LOGGER.info(u'Fetching %s…', url)
    return requests.get(url)


def ftr_process(url=None, content=None, config=None):
    u""" process an URL, or some already fetched content from a given URL.

    :param url: The URL of article to extract. Can be
        ``None``, but only if you provide both ``content`` and
        ``config`` parameters.
    :type url: str, unicode or ``None``

    :param content: the HTML content already downloaded. If given,
        it will be used for extraction, and the ``url`` parameter will
        be used only for site config lookup if ``config`` is not given.
        Please, only ``unicode`` to avoid charset errors.
    :type content: unicode or ``None``

    :param config: if ``None``, it will be looked up from ``url`` with as
        much love and AI as possible. But don't expect too much.
    :type config: a :class:`SiteConfig` instance or ``None``


    :raises:
        - :class:`RuntimeError` in all parameters-incompatible situations.
          Please RFTD carefully, and report strange unicornic edge-cases.
        - :class:`SiteConfigNotFound` if no five-filter site config can
          be found.
        - any raw ``requests.*`` exception, network related, if anything
          goes wrong during url fetching.

    :returns:
        - either a :class:`ContentExtractor` instance with extracted
          (and :attr:`.failures`) attributes set, in case a site config
          could be found.
          When the extractor knows how to handle multiple-pages articles,
          all pages contents will be extracted and cleaned — if relevant —
          and concatenated into the instance :attr:`body` attribute.
          The :attr:`next_page_link` attribute will be a ``list``
          containing all sub-pages links. Note: the first link is the one
          you fed the extractor with ; it will not be repeated in the list.
        - or ``None``, if content was not given and url fetching returned
          a non-OK HTTP code, or if no site config could be found (in that
          particular case, no extraction at all is performed).
    """

    if url is None and content is None and config is None:
        raise RuntimeError('At least one of url or the couple content/config '
                           'argument must be present.')

    if content is not None and url is None and config is None:
        raise RuntimeError('Passing content only will not give any result.')

    if content is None:
        if url is None:
            raise RuntimeError('When content is unset, url must be set.')

        try:
            result = requests_get(url)

            if result.status_code != requests.codes.ok:
                LOGGER.error(u'Wrong status code in return while getting “%s”',
                             url)
                return None

            # Override before accessing result.text ; see `requests` doc.
            result.encoding = detect_encoding_from_requests_response(result)

            LOGGER.info(u'Downloaded %s bytes as %s text.',
                        len(result.text), result.encoding)

            # result.text is always unicode
            content = result.text

        except:
            LOGGER.error(u'Content could not be fetched from URL %s', url)
            raise

    if config is None:
        # This can eventually raise SiteConfigNotFound
        config = SiteConfig(site_config_text=ftr_get_config(url))

    extractor = ContentExtractor(config)

    if extractor.process(html=content):

        # This is recursive. Yeah.
        if extractor.next_page_link is not None:
            next_extractor = ftr_process(url=extractor.next_page_link)
            extractor.body += next_extractor.body

            extractor.next_page_link = [extractor.next_page_link]

            if next_extractor.next_page_link is not None:
                extractor.next_page_link.extend(next_extractor.next_page_link)

        return extractor

    return None



# -*- coding: utf-8 -*-
u""" Python FTR content extractor class and utils.

.. Copyright 2015 Olivier Cortès <oc@1flow.io>.

    This file is part of the python-ftr project.

    python-ftr is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    python-ftr is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public
    License along with python-ftr. If not, see http://www.gnu.org/licenses/
"""
import logging
try:
    from lxml import etree
    # from lxml.cssselect import CSSSelector
    from readability.readability import Document

except ImportError:
    # Avoid a crash during setup.py
    # In normal conditions where deps are installed, this should not happen.
    # Yeah I know it's an evil hack.
    pass

from StringIO import StringIO

LOGGER = logging.getLogger(__name__)


try:
    import tidylib

except:
    LOGGER.warning('Please install the tidy library (eg. tidyhtml on Arch '
                   'Linux or libtidy-0.99-0 on Debian/Ubuntu). Things will '
                   'work without it, but results will be eventually better '
                   u'with it.')

    tidylib = None


class ContentExtractor(object):

    """
    Extract content from HTML, using patterns specified in site configuration.

    .. todo:: complete autodection port. As `1flow <http://1flow.io/>`_
        has its own autodetection code and FTR was primarily developed to
        be integrated in a complex parsing chain, I didn't reimplement
        all of the automatic extraction code here.

    Original PHP was version 1.0, written on 2013-02-05 by Keyvan Minoukadeh.
    """

    tidy_config = {
        'clean': True,
        'indent': 0,
        'output-xhtml': True,
        'logical-emphasis': True,
        'show-body-only': False,
        'new-blocklevel-tags':
            'article, aside, footer, header, hgroup, menu, '
            'nav, section, details, datagrid',
        'new-inline-tags': 'mark, time, meter, progress, data',
        'wrap': 0,
        'drop-empty-paras': True,
        'drop-proprietary-attributes': False,
        'enclose-text': True,
        'enclose-block-text': True,
        'merge-divs': True,
        'merge-spans': True,
        'char-encoding': 'utf8',
        'hide-comments': True
    }

    def __init__(self, config):
        """ Hello my dear pep257. This is an init, bright and shiny. """

        self.reset()

        self.config = config
        # LOGGER.info(u'Set config to %s.', config)

    def reset(self):
        """ (re)set all instance attributes to default.

        Every attribute is set to ``None``, except :attr:`author`
        and :attr:`failures` which are set to ``[]``.
        """

        self.config = None
        self.html = None
        self.parsed_tree = None
        self.tidied = False
        self.next_page_link = None
        self.title = None
        self.author = set()
        self.language = None
        self.date = None
        self.body = None
        self.failures = set()
        self.success = False

        LOGGER.debug(u'Reset extractor instance to defaults/empty.')

    def _process_replacements(self, html):
        """ Do raw string replacements on :param:`html`. """

        if self.config.find_string:
            for find_pattern, replace_pattern in self.config.replace_patterns:
                html = html.replace(find_pattern, replace_pattern)

            LOGGER.info(u'Done replacements.')

        return html

    def _tidy(self, html, smart_tidy):
        """ Tidy HTML if we have a tidy method.

        This fixes problems with some sites which would otherwise trouble
        DOMDocument's HTML parsing.

        Although sometimes it makes the problem worse, which is why we can
        override it in site config files.
        """

        if self.config.tidy and tidylib and smart_tidy:

            try:
                document, errors = tidylib.tidy_document(html, self.tidy_config)

            except UnicodeDecodeError:
                # For some reason, pytidylib fails to decode, whereas the
                # original html content converts perfectly manually.
                document, errors = tidylib.tidy_document(html.encode('utf-8'),
                                                         self.tidy_config)
                document = document.decode('utf-8')
            # if errors:
            #     LOGGER.debug(u'Ignored errors returned by tidylib: %s',
            #                  errors)

            self.tidied = True
            self.html = document

            LOGGER.info(u'Tidied document.')

        else:
            self.html = html

    def _parse_html(self):
        """ Load the parser and parse `self.html`. """

        if self.config.parser != 'lxml':
            raise NotImplementedError('%s parser not implemented' %
                                      self.config.parser)

        self.parser = etree.HTMLParser()
        self.parsed_tree = etree.parse(StringIO(self.html), self.parser)

        # Wanna use CSS selector?
        #
        # td_empformbody = CSSSelector('td.empformbody')
        # for elem in td_empformbody(tree):
        #     # Do something with these table cells.

    def _extract_next_page_link(self):
        """ Try to get next page link. """

        # HEADS UP: we do not abort if next_page_link is already set:
        #           we try to find next (eg. find 3 if already at page 2).

        for pattern in self.config.next_page_link:
            items = self.parsed_tree.xpath(pattern)

            if not items:
                continue

            if len(items) == 1:
                item = items[0]

                if 'href' in item.keys():
                    self.next_page_link = item.values()[0]

                else:
                    self.next_page_link = item.text.strip()

                # First found link is the good one.
                break

            else:
                LOGGER.warning(u'%s items for next-page link %s',
                               items, pattern)

    def _extract_title(self):
        """ Extract the title and remove it from the document.

        If title has already been extracted, this method will do nothing.

        If removal cannot happen or fails, the document is left untouched.
        """

        if self.title:
            return

        for pattern in self.config.title:
            items = self.parsed_tree.xpath(pattern)

            if not items:
                continue

            if isinstance(items, basestring):
                # In case xpath returns only one element.
                items = [items]

            if len(items) == 1:
                item = items[0]

                try:
                    self.title = item.text

                except AttributeError:
                    # '_ElementStringResult' object has no attribute 'text'
                    self.title = unicode(item)

                LOGGER.info(u'title set to “%s”', self.title)

                try:
                    item.getparent().remove(item)

                except TypeError:
                    # Argument 'element' has incorrect type (expected
                    # lxml.etree._Element, got _ElementStringResult)
                    pass

                except:
                    LOGGER.exception(u'Could not remove title from document.')

                # Exit at first item found.
                break

            else:
                LOGGER.warning(u'Multiple items (%s) for title pattern %s.',
                               items, pattern)

    def _extract_author(self):
        """ Extract author(s) if not already done. """

        if bool(self.author):
            return

        for pattern in self.config.author:

            items = self.parsed_tree.xpath(pattern)

            if isinstance(items, basestring):
                # In case xpath returns only one element.
                items = [items]

            for item in items:

                if isinstance(item, basestring):
                    # '_ElementStringResult' object has no attribute 'text'
                    stripped_author = unicode(item).strip()

                else:
                    try:
                        stripped_author = item.text.strip()

                    except AttributeError:
                        # We got a <div>…
                        stripped_author = etree.tostring(item)

                if stripped_author:
                    self.author.add(stripped_author)
                    LOGGER.info(u'Author found: %s', stripped_author)

    def _extract_language(self):
        """ Extract language from the HTML ``<head>`` tags. """

        if self.language:
            return

        found = False

        for pattern in (
            '//html[@lang]/@lang',
            '//meta[@name="DC.language"]/@content',
        ):
            for item in self.parsed_tree.xpath(pattern):
                stripped_language = item.strip()

                if stripped_language:
                    self.language = stripped_language
                    LOGGER.info(u'Language found: %s', stripped_language)
                    found = True
                    break

            if found:
                break

    def _extract_date(self):
        """ Extract date from HTML. """

        if self.date:
            return

        found = False

        for pattern in self.config.date:

            items = self.parsed_tree.xpath(pattern)

            if isinstance(items, basestring):
                # In case xpath returns only one element.
                items = [items]

            for item in items:
                if isinstance(item, basestring):
                    # '_ElementStringResult' object has no attribute 'text'
                    stripped_date = unicode(item).strip()

                else:
                    try:
                        stripped_date = item.text.strip()

                    except AttributeError:
                        # .text is None. We got a <div> item with span-only
                        # content. The result will probably be completely
                        # useless to a python developer, but at least we
                        # didn't fail handling the siteconfig directive.
                        stripped_date = etree.tostring(item)

                if stripped_date:
                    # self.date = strtotime(trim(elems, "; \t\n\r\0\x0B"))
                    self.date = stripped_date
                    LOGGER.info(u'Date found: %s', stripped_date)
                    found = True
                    break

            if found:
                break

    def _strip_unwanted_elements(self):

        def _remove(xpath_expression):
            for item in self.parsed_tree.xpath(xpath_expression):
                item.getparent().remove(item)
                LOGGER.debug(u'Removed unwanted item %s', item)

        # Strip elements that use xpath expressions.
        for pattern in self.config.strip:
            _remove(pattern)

        # Strip elements using #id or .class attribute values.
        for pattern in self.config.strip_id_or_class:
            _remove(
                "//*[contains(@class, '{0}') or contains(@id, '{0}')]".format(
                    pattern.replace('"', '').replace("'", '')
                )
            )

        # Strip images using src attribute values.

        for pattern in self.config.strip_image_src:
            _remove(
                "//img[contains(@src, '{0}')]".format(
                    pattern.replace('"', '').replace("'", '')
                )
            )

        # Strip elements using Readability.com and Instapaper.com ignore
        # classes names .entry-unrelated and .instapaper_ignore
        # See https://www.readability.com/publishers/guidelines/#view-plainGuidelines  # NOQA
        # and http://blog.instapaper.com/post/730281947 for details.
        _remove(
            "//*[contains(concat(' ',normalize-space(@class),' ')"
            ",' entry-unrelated ') or contains(concat(' ',"
            "normalize-space(@class),' '),' instapaper_ignore ')]"
        )

        # Strip elements that contain style="display: none;".
        _remove("//*[contains(@style,'display:none')]")

    def _extract_body(self):
        """ Extract the body content from HTML. """

        def is_descendant_node(parent, node):
            node = node.getparent()
            while node is not None:
                if node == parent:
                    return True
                node = node.getparent()
            return False

        for pattern in self.config.body:
            items = self.parsed_tree.xpath(pattern)

            if len(items) == 1:
                if self.config.prune:
                    self.body = Document(etree.tostring(items[0])).summary()

                else:
                    self.body = etree.tostring(items[0])

                # We've got a body now.
                break

            else:
                appended_something = False
                body = etree.Element("root")

                for item in items:
                    if item.getparent() is None:
                        continue

                    is_descendant = False

                    for parent in body:
                        if (is_descendant_node(parent, item)):
                            is_descendant = True
                            break

                    if not is_descendant:

                        if self.config.prune:

                            # Clean with readability. Needs
                            # to-string conversion first.
                            pruned_string = Document(
                                etree.tostring(item)).summary()

                            # Re-parse the readability string
                            # output and include it in our body.
                            new_tree = etree.parse(
                                StringIO(pruned_string), self.parser)

                            try:
                                body.append(
                                    new_tree.xpath('//html/body/div/div')[0]
                                )
                            except IndexError:
                                LOGGER.error(u'Pruning this item did not '
                                             u'work:\n\n%s\n\nWe got: “%s” '
                                             u'and skipped it.',
                                             etree.tostring(
                                                 item).replace(u'\n', u''),
                                             pruned_string.replace(u'\n', u''))
                                pass

                        else:
                            etree.SubElement(body, item)

                        appended_something = True

                if appended_something:
                    self.body = etree.tostring(body)

                    # We've got a body now.
                    break

    def _auto_extract_if_failed(self):
        """ Try to automatically extract as much as possible. """

        if not self.config.autodetect_on_failure:
            return

        readabilitized = Document(self.html)

        if self.title is None:
            if bool(self.config.title):
                self.failures.add('title')

            title = readabilitized.title().strip()

            if title:
                self.title = title
                LOGGER.info(u'Got a title in automatic mode.')

            else:
                self.failures.add('title')

        if self.body is None:
            if bool(self.config.body):
                self.failures.add('body')

            body = readabilitized.summary().strip()

            if body:
                self.body = body
                LOGGER.info(u'Extracted a body in automatic mode.')

            else:
                self.failures.add('body')

        for attr_name in ('date', 'language', 'author', ):
            if not bool(getattr(self, attr_name, None)):
                if bool(getattr(self.config, attr_name, None)):
                    self.failures.add(attr_name)
                    LOGGER.warning(u'Could not extract any %s from %s.',
                                   attr_name, getattr(self.config, attr_name))
                    # import ipdb; ipdb.set_trace()

    def process(self, html, url=None, smart_tidy=True):
        u""" Process HTML content or URL.

        For automatic extraction patterns and cleanups, :mod:`readability-lxml`
        is used, to stick as much as possible to the original PHP
        implementation and produce at least similar results with the same
        site config on the same article/content.

        :param html: an unicode string containing a full HTML page content.
            Expected to have a ``DOCTYPE`` and all other standard
            attributes ; eg. HTML fragments are not supported.
            It will be replaced, tidied, cleaned, striped, and all
            metadata and body attributes will be extracted from it.
            Beware : this HTML piece will be mauled. See source code for
            exact processing workflow, it's quite gorgeous.
        :type html: unicode

        :param url: as of version 0.5, this parameter is ignored. (**TODO**)
        :type url: str, unicode or ``None``

        :param smart_tidy: When ``True`` (default), runs :mod:`pytidylib`
            to tidy the HTML, after after run ``find_string``/``replace_string``
            replacements and before running extractions.
        :type smart_tidy: bool

        :returns: ``True`` on success, ``False`` on failure.
        :raises:
            - :class:`RuntimeError` if config has not been set at
              instantiation. This should change in the future by looking
              up a config if an ``url`` is passed as argument.

        .. note:: If tidy is used and no result is produced, we will try
            again without tidying.
            Generally speaking, tidy helps us deal with PHP's patchy HTML
            parsing (LOOOOOL. Zeriously?) most of the time but it has
            problems of its own which we try to avoid with this option.
            In the Python implementation, `pytidylib` has showed to help
            sanitize a lot the HTML before processing it. But nobody's
            perfect, and errors can happen in the Python world too, thus
            the *tidy* behavior was thought sane enough to be keep.
        """

        # TODO: re-implement URL handling with self.reset() here.

        if self.config is None:
            raise RuntimeError(u'extractor site config is not set.')

        # TODO: If re-running ourselves over an already-replaced string,
        #       this should just do nothing because everything has been
        #       done. We should have a test for that.
        html = self._process_replacements(html)

        # We keep the html untouched after replacements.
        # All processing happens on self.html after this point.
        self._tidy(html, smart_tidy)

        # return

        self._parse_html()

        self._extract_next_page_link()

        self._extract_title()

        self._extract_author()

        self._extract_language()

        self._extract_date()

        self._strip_unwanted_elements()

        self._extract_body()

        # TODO: re-implement auto-detection here.
        # NOTE: hNews extractor was here.
        # NOTE: instapaper extractor was here.

        self._auto_extract_if_failed()

        if self.title is not None or self.body is not None \
            or bool(self.author) or self.date is not None \
                or self.language is not None:
            self.success = True

        # if we've had no success and we've used tidy, there's a chance
        # that tidy has messed up. So let's try again without tidy...
        if not self.success and self.tidied and smart_tidy:
            self.process(html, url=None, smart_tidy=False)

        return self.success



# -*- coding: utf-8 -*-
u""" Python FTR configuration class and utils.

The configuration files are named “ siteconfig ” in Five Filters
terminology, and probably stands for “ website configuration ” sets.

A `siteconfig` is a simple text file with a ``key: value`` format.
These files are loaded into :class:`SiteConfig` instances. They are
named after the website they hold rules for.

.. Copyright 2015 Olivier Cortès <oc@1flow.io>.

    This file is part of the python-ftr project.

    python-ftr is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    python-ftr is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public
    License along with python-ftr. If not, see http://www.gnu.org/licenses/
"""
import os
import re
import logging

LOGGER = logging.getLogger(__name__)

try:
    from ordered_set import OrderedSet

except ImportError:
    # Avoid a crash during setup.py
    # In normal conditions where deps are installed, this should not happen.
    # Yeah I know it's an evil hack.
    pass

try:
    from cacheops import cached

except Exception, e:
    LOGGER.warning(u'Cacheops seems not installed or not importable '
                   u'(exception was: %s). Running without cache.', e)
    from functools import wraps

    def cached(*a, **kw):
        """ A no-op decorator in case :mod:`cacheops` is not installed. """

        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return decorator


# defaults to 3 days of caching for website configuration
CACHE_TIMEOUT = int(os.environ.get('PYTHON_FTR_CACHE_TIMEOUT', 345600))

HOSTNAME_REGEX = re.compile(
    r'/^(([a-z0-9-]*[a-z0-9])\.)*([a-z0-9-]*[a-z0-9])$/',
    re.IGNORECASE | re.UNICODE
)


class SiteConfigException(Exception):

    """ Abstract base class for all :class:`SiteConfig` related exceptions.

    Never raised directly, but you can catch it in an `except` block to
    get them all at once.
    """

    pass


class SiteConfigNotFound(SiteConfigException):

    """ Raised when no site config can be found for a given website or url. """

    pass


class InvalidSiteConfig(SiteConfigException):

    """ Raised when an unrecoverable error is found in a site config.

    Eg. when it has a very bad syntax error or a missing mandatory directive.

    As of version 0.5, all directives are optional, thus this exception
    is raised only in case of a non-matching number of ``find_string``
    / ``replace_string`` pairs. This could change in the future.
    """

    pass


@cached(timeout=CACHE_TIMEOUT)
def ftr_get_config(website_url, exact_host_match=False):
    """ Download the Five Filters config from centralized repositories.

    The first entry found is returned. If no configuration is found,
    `None` is returned. If :mod:`cacheops` is installed, the result will
    be cached with a default expiration delay of 3 days.

    :param exact_host_match: If ``False`` (default), we will look for
        wildcard config matches. For example if host is
        ``www.test.example.org``, we will try looking up
        ``test.example.org`` and ``example.org``.
    :param exact_host_match: bool

    :param website_url: either a full web URI (eg.
        ``http://www.website.com:PORT/path/to/a/page.html``) or simply
        a domain name (eg. ``www.website.com``). In case of a domain name,
        no check is performed yet, be careful of what you pass.
    :type website_url: str or unicode

    :returns: string -- the loaded site config, if found.
    :raises: :class:`SiteConfigNotFound` if no config could be found.

    .. note:: Whatever ``exact_host_match`` value is, the ``www`` part is
        always removed from the URL or domain name.

    .. todo:: there is currently no merging/cascading of site configs. In
        the original Five Filters implementation, primary and secondary
        configurations were merged. We could eventually re-implement this
        part if needed by someone. PRs welcome as always.
    """

    import requests
    from sparks.utils.http import split_url

    repositories = [
        x.strip() for x in os.environ.get(
            'PYTHON_FTR_REPOSITORIES',
            u'https://raw.githubusercontent.com/1flow/ftr-site-config/master/ '
            u'https://raw.githubusercontent.com/fivefilters/ftr-site-config/master/'  # NOQA
        ).split() if x.strip() != u'']

    try:
        proto, host_and_port, remaining = split_url(website_url)

    except:
        host_and_port = website_url

    host_domain_parts = host_and_port.split(u'.')

    # we don't store / use the “www.” part of domain name in siteconfig.
    if host_domain_parts[0] == u'www':
        host_domain_parts = host_domain_parts[1:]

    if exact_host_match:
        domain_names = [u'.'.join(host_domain_parts)]

    else:
        domain_names = [
            u'.'.join(host_domain_parts[-i:])
            for i in reversed(range(2, len(host_domain_parts) + 1))
        ]

    LOGGER.debug(u'Gathering configurations for domains %s from %s.',
                 domain_names, repositories)

    for repository in repositories:
        # try, in turn:
        #   website.ext.txt
        #   .website.ext.txt
        for domain_name in domain_names:

            skip_repository = False

            for full_name in (
                u'{0}{1}.txt'.format(repository, domain_name),
                u'{0}.{1}.txt'.format(repository, domain_name),
            ):
                result = requests.get(full_name)

                if result.status_code == requests.codes.ok:
                    if u'text/plain' not in result.headers.get('content-type') \
                        or u'<!DOCTYPE html>' in result.text \
                        or u'<html ' in result.text \
                            and u'</html>' in result.text:
                        LOGGER.error(u'“%s” repository URL does not return '
                                     u'RAW plain text results.')

                        skip_repository = True
                        break

                    LOGGER.info(u'Using siteconfig for domain %s from %s.',
                                domain_name, full_name)
                    return result.text

                if skip_repository:
                    break

            if skip_repository:
                break

    raise SiteConfigNotFound(
        u'No configuration found for domains {0} in repositories {1}'.format(
            u', '.join(domain_names), u', '.join(repositories)
        )
    )


def ftr_string_to_instance(config_string):
    """ Return a :class:`SiteConfig` built from a ``config_string``.

    Simple syntax errors are just plainly ignored, and logged as warnings.

    :param config_string: a full site config file, raw-loaded from storage
        with something like
        ``config_string = open('path/to/site/config.txt', 'r').read()``.
    :type config_string: str or unicode

    :returns: a :class:`SiteConfig` instance.
    :raises: :class:`InvalidSiteConfig` in case of an unrecoverable error.

    .. note:: See the source code for supported directives names.
    """

    config = SiteConfig()

    for line_number, line_content in enumerate(
            config_string.strip().split(u'\n'), start=1):

        line_content = line_content.strip()

        # Skip empty lines & comments.
        if not line_content or line_content.startswith(u'#'):
            continue

        try:
            key, value = [
                x.strip() for x in line_content.strip().split(u':', 1)
            ]

        except:
            LOGGER.warning(u'Unrecognized syntax “%s” on line #%s.',
                           line_content, line_number)
            continue

        if not key or (not value and key != 'replace_string'):
            LOGGER.warning(u'Empty key or value in “%s” on line #%s.',
                           line_content, line_number)
            continue

        # Commands for which we accept multiple statements.
        elif key in (
            'title', 'body', 'author', 'date',
            'strip', 'strip_id_or_class', 'strip_image_src',
            'single_page_link', 'single_page_link_in_feed',
            'next_page_link',
            'http_header',

            'find_string',
            'replace_string',

            'test_url',
            'test_contains',
            'test_title',
            'test_date',
            'test_author',
            'test_language',
        ):

            if key.endswith(u'_string'):
                # Append to list. Duplicites are allowed.
                getattr(config, key).append(value)

            else:
                # Add to set, preserving order but squashing duplicates.
                getattr(config, key).add(value)

        # Single statement commands that evaluate to True or False.
        elif key in ('tidy', 'prune', 'autodetect_on_failure', ):

            if value.lower() in ('no', 'false', '0', ):
                setattr(config, key, False)

            else:
                setattr(config, key, bool(value))

        # Single statement commands stored as strings.
        elif key in ('parser', ):
            setattr(config, key, value)

        # The “replace_string(………): replace_value” one-liner syntax.
        elif key.startswith('replace_string(') and key.endswith(')'):
            # These 2 are lists, not sets.
            config.find_string.append(key[15:-1])
            config.replace_string.append(value)

        else:
            LOGGER.warning(u'Unsupported directive %s = %s on line #%s.',
                           key, value, line_number)

    find_count = len(config.find_string)
    replace_count = len(config.replace_string)

    if find_count != replace_count:
        raise InvalidSiteConfig(u'find_string and remplace_string do not '
                                u'correspond ({0} != {1})'.format(
                                    find_count, replace_count))

    return config


class SiteConfig(object):

    """ Holds extraction pattern and other directives for a given website.

    See :class:`ContentExtractor` class to see how it's used.

    This is the Five Filters SiteConfig object, ported to Python.

    Original was PHP, version 0.8, written by Keyvan Minoukadeh.
    """

    defaults = {
        'tidy': True,
        'parser': 'libxml',
        'prune': True,
        'autodetect_on_failure': True,
    }

    def __unicode__(self):
        """ Print title & body. """
        return u'title: %s, body: %s' % (self.title, self.body)

    def __init__(self, host=None, site_config_text=None):
        """ Load a first config, either from a hostname or a string config. """

        self.reset()

        if host is not None:
            self.load(host)

        if site_config_text is not None:
            self.append(ftr_string_to_instance(site_config_text))

    def reset(self):
        """ (re)set all attributes to defaults (eg. empty sets or ``None``). """

        # Use first matching element as title (0 or more xpath expressions)
        self.title = OrderedSet()

        # Use first matching element as body (0 or more xpath expressions)
        self.body = OrderedSet()

        # Use first matching element as author (0 or more xpath expressions)
        self.author = OrderedSet()

        # Use first matching element as date (0 or more xpath expressions)
        self.date = OrderedSet()

        # Strip elements matching these xpath expressions (0 or more)
        self.strip = OrderedSet()

        # Strip 0 or more elements which contain these
        # strings in the id or class attribute.
        self.strip_id_or_class = OrderedSet()

        # Strip 0 or more images which contain
        # these strings in the src attribute.
        self.strip_image_src = OrderedSet()

        # Additional HTTP headers to send
        # NOT YET USED
        self.http_header = OrderedSet()

        # For those 3, None means that default will be used. But we need
        # None to distinguish from False during multiple configurations
        # merges.
        self.tidy = None
        self.prune = None
        self.autodetect_on_failure = None

        # Test URL - if present, can be used to test the config above
        self.test_url = OrderedSet()
        self.test_contains = OrderedSet()

        # Single-page link should identify a link element or URL pointing
        # to the page holding the entire article.
        #
        # This is useful for sites which split their articles across
        # multiple pages. Links to such pages tend to display the first
        # page with links to the other pages at the bottom.
        #
        # Often there is also a link to a page which displays the entire
        # article on one page (e.g. 'print view').
        #
        # `single_page_link` should be an XPath expression identifying the
        # link to that single page. If present and we find a match, we will
        # retrieve that page and the rest of the options in this config will
        # be applied to the new page.
        self.single_page_link = OrderedSet()

        self.next_page_link = OrderedSet()

        # Single-page link in feed? - same as above, but patterns applied
        # to item description HTML taken from feed. XXX
        self.single_page_link_in_feed = OrderedSet()

        # Which parser to use for turning raw HTML into a DOMDocument,
        # either `libxml` (PHP) / `lxml` (Python) or `html5lib`. Defaults
        # to `lxml` if None.
        self.parser = None

        # Strings to search for in HTML before processing begins. Goes by
        # pairs with `replace_string`. Not a set because we can have more
        # than one of the same, to be replaced by different values.
        self.find_string = []

        # Strings to replace those found in `find_string` before HTML
        # processing begins.
        self.replace_string = []

    def load(self, host, exact_host_match=False):
        """ Load a config for a hostname or url.

        This method calls :func:`ftr_get_config` and :meth`append`
        internally. Refer to their docs for details on parameters.
        """

        # Can raise a SiteConfigNotFound, intentionally bubbled.
        config = ftr_get_config(host, exact_host_match)

        if config is None:
            LOGGER.error(u'Error while loading configuration.')
            return

        self.append(ftr_string_to_instance(config))

    def append(self, newconfig):
        """ Append another site config to current instance.

        All ``newconfig`` attributes are appended one by one to ours.
        Order matters, eg. current instance values will come first when
        merging.

        Thus, if you plan to use some sort of global site config with
        more generic directives, append it last for specific directives
        to be tried first.

        .. note:: this method is also aliased to :meth:`merge`.
        """

        # Check for commands where we accept multiple statements (no test_url)
        for attr_name in (
            'title', 'body', 'author', 'date',
            'strip', 'strip_id_or_class', 'strip_image_src',
            'single_page_link', 'single_page_link_in_feed',
            'next_page_link', 'http_header'
        ):
            # Append to ordered set. We keep ordering, but no duplicates.
            current_set = getattr(self, attr_name)
            for val in getattr(newconfig, attr_name):
                # Too bad ordered set has no .union() method.
                current_set.add(val)
            setattr(self, attr_name, current_set)

        # Check for single statement commands;
        # we do not overwrite existing values.
        for attr_name in (
            'parser', 'tidy', 'prune', 'autodetect_on_failure'
        ):
            if getattr(self, attr_name) is None:
                if getattr(newconfig, attr_name) is None:
                    setattr(self, attr_name, self.defaults[attr_name])
                else:
                    setattr(self, attr_name, getattr(newconfig, attr_name))

        # HEADS UP: PHP → Python port.
        if self.parser == 'libxml':
            self.parser = 'lxml'

        for attr_name in ('find_string', 'replace_string', ):
            # Find/replace strings are lists, we extend.
            getattr(self, attr_name).extend(getattr(newconfig, attr_name))

        if self.find_string:
            # This will ease replacements in the extractor.
            self.replace_patterns = zip(
                self.find_string, self.replace_string)

        else:
            self.replace_patterns = None

    # method aliasing for API compatibility.
    merge = append



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Python Five-Filters setup.py. """

import sys

from setuptools import setup, find_packages

if '.' in sys.path:
    sys.path.remove('.')

# We want to be sure that Python will import the sparks from here, and not
# the one eventually installed system-wide or in the current virtualenv.
sys.path.insert(0, '.')

from ftr.version import version

setup(
    name="ftr",
    version=version,
    author="Olivier Cortès",
    author_email="contact@oliviercortes.com",
    description="HTML Article cleaner / extractor, Five-Filters compatible.",
    long_description=open('README.md').read(),
    url="https://github.com/1flow/python-ftr",
    packages=find_packages(),
    include_package_data=True,
    dependency_links=[
        'https://github.com/1flow/sparks/tarball/master#egg=sparks',
        'https://github.com/Karmak23/humanize/tarball/master#egg=humanize',
    ],
    install_requires=[
        # 'html5lib',
        'pytidylib',
        'lxml',
        'ordered-set',
        'readability-lxml',
        'requests',
    ],
    extras_require={
        'cache':  ['cacheops'],
    },
    keywords=(
        'parsing',
        'websites',
        'articles',
    ),
    license='AGPLv3',
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",  # NOQA
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Topic :: Text Processing :: Filters",
    ],
)



# -*- coding: utf-8 -*-
u"""
Copyright 2015 Olivier Cortès <oc@1flow.io>.

This file is part of the python-ftr project.

python-ftr is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

python-ftr is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public
License along with python-ftr. If not, see http://www.gnu.org/licenses/


"""

# Import the sparks-managed version.
from .version import version  # NOQA

from .config import (  # NOQA
    ftr_get_config as get_config,
    SiteConfig,
    SiteConfigException,
    SiteConfigNotFound,
    InvalidSiteConfig,
)

from .extractor import (  # NOQA
    ContentExtractor
)

from process import (  # NOQA
    ftr_process as process,
)

# Advertise version to external tools like Sentry.
__version__ = version
